#include <stdlib.h>
#include <stdarg.h>

#include "../Daodan.h"
#include "Input.h"
#include "../Oni/Oni.h"
#include "../Daodan_Config.h"
#include "../Daodan_Patch.h"
#include "Utility.h"

#define EVENT_KEYPRESS_SECURETIME 3

typedef struct {
	uint32_t key;
	const char* actionname;
	ActionEventType_t eventType;
	uint32_t keydownTimeoutTicks;
	int64_t lastevent;

	CustomActionCallback_t callback;
	CustomActionCallbackArgument callbackArgument;
} DD_CustomAction_t;

static DD_CustomAction_t customActions[100];


void Input_RegisterCustomAction (const char* actionname, ActionEventType_t eventType, uint32_t keydownTimeoutTicks, CustomActionCallback_t callback, CustomActionCallbackArgument callbackArgument) {
	uint16_t i = 0;
	DD_CustomAction_t* cur = customActions;

	while ( (i < ARRAY_SIZE(customActions)) && (cur->callback != 0)) {
		cur++;
		i++;
	}

	if (i < ARRAY_SIZE(customActions)) {
		cur->actionname = actionname;
		cur->eventType = eventType;
		cur->keydownTimeoutTicks = keydownTimeoutTicks;
		cur->callback = callback;
		cur->callbackArgument = callbackArgument;
	} else {
		STARTUPMESSAGE("Registering action %s failed, maximum actions reached", actionname);
	}
}


_LIrBinding_Add Oni_LIrBinding_Add = (_LIrBinding_Add)0;
uint16_t ONICALL DD_LIrBinding_Add(uint32_t key, const char* name) {
	DD_CustomAction_t* cur;
	for (cur = customActions; cur->callback != 0; cur++) {
		if (!strcmp(name, cur->actionname)) {
			cur->key = key;
			return 0;
		}
	}

	return Oni_LIrBinding_Add(key, name);
}

_LIrActionBuffer_Add Oni_LIrActionBuffer_Add = (_LIrActionBuffer_Add)0;
void ONICALL DD_LIrActionBuffer_Add(void* unknown, LItDeviceInput* input) {
	DD_CustomAction_t* cur;
	for (cur = customActions; cur->callback != 0; cur++) {
		if (cur->key == input->input) {
			int64_t curTime = UUrMachineTime_Sixtieths();
			if (cur->eventType == EVENT_KEYPRESS) {
				if (cur->lastevent + EVENT_KEYPRESS_SECURETIME < curTime) {
					cur->callback(cur->callbackArgument);
				}
				cur->lastevent = curTime;
			} else if (cur->eventType == EVENT_KEYDOWN) {
				if (cur->lastevent + cur->keydownTimeoutTicks < curTime) {
					cur->callback(cur->callbackArgument);
					cur->lastevent = curTime;
				}
			}
			return;
		}
	}
	Oni_LIrActionBuffer_Add(unknown, input);
}

void Input_PatchCode () {
	Oni_LIrBinding_Add = DDrPatch_MakeDetour((void*)LIrBinding_Add, (void*)DD_LIrBinding_Add);
	Oni_LIrActionBuffer_Add = DDrPatch_MakeDetour((void*)LIrActionBuffer_Add, (void*)DD_LIrActionBuffer_Add);
}

// Daodan Input (here for now)

// Set in Patches.c if the Daodan input patches are applied. This just enables
// the windows message handling for now
bool DDgUseDaodanInput = false;

// Multiplier for mouse values
static float DDgMouseSensitivity = 1.0;

// Temporary action buffer that we build over the duration of a frame with the
// input we're going to send to Oni. This includes the accumulated relative
// movement of the mouse cursor and all actions (keyboard and mouse buttons)
// that were pressed this frame (but not held down from previous frames - that
// gets added later.)
static LItActionBuffer DDgActionBuffer = { 0 };

// Temporary buffer containing the current state of the keyboard and mouse
// buttons, that is, if they're being held now
static char DDgInputState[256] = { 0 };

static void CenterCursor(void)
{
	// This can be set to false by script. Not sure why you'd turn it off, but
	// let's respect it.
	if (!LIgCenterCursor)
		return;

	RECT rc;
	if (!GetClientRect(LIgPlatform_HWND, &rc))
		return;
	POINT mid = { rc.right / 2, rc.bottom / 2 };
	if (!ClientToScreen(LIgPlatform_HWND, &mid))
		return;
	SetCursorPos(mid.x, mid.y);
}

void ONICALL DDrPlatform_Mode_Set(int active)
{
	// Oni's input system uses LIgPlatform_HWND instead of
	// ONgPlatformData.Window, but they should both have the same value
	DDmAssert(LIgPlatform_HWND);

	// Clear the input state when switching input modes
	for (int i = 0; i < ARRAY_SIZE(DDgInputState); i++)
		DDgInputState[i] = 0;
	DDgActionBuffer = (LItActionBuffer) { 0 };

	// Center the cursor before switching modes. Raw Input doesn't need the
	// cursor to be centered, but when switching modes, centering the cursor
	// means it will be in a predictable position for using the pause or F1
	// menu, which are centered on the screen. Also, the cursor must be inside
	// the clip region for ClipCursor to work later on.
	CenterCursor();

	// If leaving input mode (switching from gameplay to menus,) unregister the
	// input device. Otherwise, register it.
	RegisterRawInputDevices(&(RAWINPUTDEVICE) {
		.usUsagePage = 0x01, // HID_USAGE_PAGE_GENERIC
		.usUsage = 0x02, // HID_USAGE_GENERIC_MOUSE
		.hwndTarget = LIgPlatform_HWND,
		.dwFlags = active ? 0 : RIDEV_REMOVE,
	}, 1, sizeof(RAWINPUTDEVICE));

	// Clip the cursor to the window bounds when entering input mode to prevent
	// other programs being clicked in windowed mode
	if (active) {
		DDgMouseSensitivity =
			DDrConfig_GetOptOfType("windows.mousesensitivity", C_FLOAT)->value.floatVal;

		RECT rc;
		if (GetClientRect(LIgPlatform_HWND, &rc)) {
			if (MapWindowRect(LIgPlatform_HWND, NULL, &rc))
				ClipCursor(&rc);
		}
	} else {
		ClipCursor(NULL);
	}
}

void ONICALL DDrPlatform_InputEvent_GetMouse(int active, LItInputEvent *event)
{
	POINT pt;
	if (!GetCursorPos(&pt))
		goto error;

	// Unlike Oni's version of this function, we support windowed mode by
	// mapping the cursor coordinates to the window's client area
	if (!ScreenToClient(LIgPlatform_HWND, &pt))
		goto error;

	*event = (LItInputEvent) { .mouse_x = pt.x, .mouse_y = pt.y };
	return;

error:
	*event = (LItInputEvent) { 0 };
	return;
}

UUtBool ONICALL DDrPlatform_TestKey(int ch, int active)
{
	// DDrPlatform_TestKey is always called with active = LIgMode_Internal

	if (active) {
		// The input system is running, which means DDgInputState will be up to
		// date, so just use that
		return DDgInputState[ch];
	} else {
		// Use Oni's map from key codes to DirectInput scan codes to get the
		// scan code we want to test for
		int sc = 0;
		for (int i = 0; i < 256; i++) {
			if (LIgPlatform_ScanCodeToChar[i] == ch) {
				sc = LIgPlatform_ScanCodeToChar[i];
				break;
			}
		}
		if (!sc)
			return UUcFalse;

		// DirectInput scan codes have 0x80 set for extended keys. Replace this
		// with an 0xe0 prefix for MapVirtualKey.
		if (sc & 0x80) {
			sc &= 0x7f;
			sc |= 0xe000;
		}
		int vkey = MapVirtualKeyA(sc, MAPVK_VSC_TO_VK_EX);

		// Now check if the key is down. We must use GetAsyncKeyState here
		// because DDrPlatform_TestKey can be called from anywhere, even before
		// we have a message loop or game loop. For example, it's called from
		// ONiMain to test the state of the shift key on startup.
		return (GetAsyncKeyState(vkey) & 0x8000) ? UUcTrue : UUcFalse;
	}
}

static void DDrActionBuffer_Add(LItActionBuffer *buf, LItDeviceInput *input)
{
	// Basically identical to Oni's LIrActionBuffer_Add at this point (except
	// for the DDmAsserts)

	// It should probably call DD_LIrActionBuffer_Add...

	LItActionDescription* descr = NULL;
	for (int i = 0; i < 100; i++) {
		if (LIgBindingArray[i].key == input->input) {
			descr = LIgBindingArray[i].descr;
			break;
		}
	}
	if (!descr)
		return;

	switch (descr->type) {
	case 1:
		DDmAssert(descr->index >= 0 && descr->index < 64);
		buf->button_flags |= 1ull << descr->index;
		break;
	case 2: case 3: case 4:
		DDmAssert(descr->index >= 0 && descr->index < 10);
		buf->analogs[descr->index] += input->analog;
		break;
	case 5:
		DDmAssert(descr->index >= 0 && descr->index < 10);
		buf->analogs[descr->index] -= input->analog;
		break;
	}
}

// Update DDgInputState and DDgActionBuffer with a new key state
static void SetKeyState(int key, bool down)
{
	DDgInputState[key] = down;

	if (down) {
		DDrActionBuffer_Add(&DDgActionBuffer, &(LItDeviceInput) {
			.input = key,
			.analog = 1.0,
		});
	}
}

static void ProcessRawInputPacket(RAWINPUT *ri)
{
	if (ri->header.dwType != RIM_TYPEMOUSE)
		return;

	// We don't handle MOUSE_MOVE_ABSOLUTE at all yet
	if (!(ri->data.mouse.usFlags & MOUSE_MOVE_ABSOLUTE)) {
		DDrActionBuffer_Add(&DDgActionBuffer, &(LItDeviceInput) {
			.input = 5,
			.analog = (float)ri->data.mouse.lLastX * 0.25 * DDgMouseSensitivity,
		});
		DDrActionBuffer_Add(&DDgActionBuffer, &(LItDeviceInput) {
			.input = 6,
			.analog = (float)ri->data.mouse.lLastY *
			          (LIgMode_InvertMouse ? -0.25 : 0.25) * DDgMouseSensitivity,
		});
	}

	// This probably doesn't obey SM_SWAPBUTTON... should it?
	if (ri->data.mouse.usButtonFlags & RI_MOUSE_LEFT_BUTTON_DOWN)
		SetKeyState(1, true);
	if (ri->data.mouse.usButtonFlags & RI_MOUSE_LEFT_BUTTON_UP)
		SetKeyState(1, false);
	if (ri->data.mouse.usButtonFlags & RI_MOUSE_RIGHT_BUTTON_DOWN)
		SetKeyState(2, true);
	if (ri->data.mouse.usButtonFlags & RI_MOUSE_RIGHT_BUTTON_UP)
		SetKeyState(2, false);
	if (ri->data.mouse.usButtonFlags & RI_MOUSE_MIDDLE_BUTTON_DOWN)
		SetKeyState(3, true);
	if (ri->data.mouse.usButtonFlags & RI_MOUSE_MIDDLE_BUTTON_UP)
		SetKeyState(3, false);

	// DirectInput seems to support four mouse buttons and Oni's DirectInput
	// code appears to use them all, so we do too with Raw Input. I haven't
	// tested to see if this actually does anything yet.
	if (ri->data.mouse.usButtonFlags & RI_MOUSE_BUTTON_4_DOWN)
		SetKeyState(4, true);
	if (ri->data.mouse.usButtonFlags & RI_MOUSE_BUTTON_4_UP)
		SetKeyState(4, false);
}

static void DrainRawInput(void)
{
	if (!LIgMode_Internal)
		return;

	UINT ri_size = 10240;
	static RAWINPUT *ri_buf = NULL;
	if (!ri_buf)
		ri_buf = calloc(1, ri_size);

	BOOL wow_hack;
	IsWow64Process(GetCurrentProcess(), &wow_hack);

	for (;;) {
		UINT count = GetRawInputBuffer(ri_buf, &ri_size, sizeof ri_buf->header);
		if (count == 0 || count == (UINT)-1)
			return;

		RAWINPUT *ri = ri_buf;
		for (UINT i = 0; i < count; i++) {
			// In WOW64, these structures are aligned like in Win64 and they
			// have to be fixed to use from 32-bit code. Yes, really.
			if (wow_hack) {
				memmove(&ri->data, ((char *)&ri->data) + 8,
					ri->header.dwSize - offsetof(RAWINPUT, data) - 8);
			}

			ProcessRawInputPacket(ri);
			ri = NEXTRAWINPUTBLOCK(ri);
		}
	}
}

UUtBool ONICALL DDiPlatform_InputEvent_GetEvent(void)
{
	// Center the cursor just in case. Raw Input doesn't need it, but sometimes
	// ClipCursor doesn't work for some reason and in that case we should still
	// prevent the user from accidentally clicking on other windows.
	if (LIgMode_Internal)
		CenterCursor();

	// Do a buffered read of raw input. Apparently this is faster for high-res
	// mice. Note that we still have to handle WM_INPUT in our wndproc in case
	// a WM_INPUT message arrives during the message loop.
	DrainRawInput();

	// Oni only processes a maximum of three messages here (for performance
	// reasons?) We're actually using Windows messages for input so we need to
	// process all of them.
	MSG msg;
	while (PeekMessageA(&msg, NULL, 0, 0, PM_REMOVE)) {
		TranslateMessage(&msg);
		DispatchMessageA(&msg);
	}

	// Oni returns true here if there are still messages to process, so that
	// LIrMode_Set and LIrMode_Set_Internal can call this repeatedly to drain
	// all messages. We always drain all messages so return false.
	return UUcFalse;
}

static bool HandleWmInput(HRAWINPUT hri, WPARAM wparam)
{
	if (!LIgMode_Internal)
		return false;

	static RAWINPUT* ri = NULL;
	static UINT ri_size = 0;
	UINT minsize = 0;

	GetRawInputData(hri, RID_INPUT, NULL, &minsize, sizeof ri->header);
	if (ri_size < minsize) {
		if (ri)
			free(ri);
		ri_size = minsize;
		ri = calloc(1, ri_size);
	}
	if (GetRawInputData(hri, RID_INPUT, ri, &ri_size, sizeof ri->header) == (UINT)-1)
		return false;

	ProcessRawInputPacket(ri);
	return true;
}

static void HandleWmWindowPosChanged(WINDOWPOS *pos)
{
	if (!LIgMode_Internal)
		return;

	CenterCursor();

	RECT rc = { pos->x, pos->y, pos->x + pos->cx, pos->y + pos->cy };
	ClipCursor(&rc);
}

static void HandleWmKeyboard(int vkey, WORD repeat_count, WORD flags)
{
	if (!LIgMode_Internal)
		return;

	bool is_extended = flags & KF_EXTENDED;
	bool is_repeat = flags & KF_REPEAT;
	bool is_up = flags & KF_UP;
	BYTE sc = LOBYTE(flags);

	// Ignore key down messages sent because of key-repeat
	if (!is_up && is_repeat)
		return;

	// DirectInput scan codes have 0x80 set for extended keys, and we're using
	// Oni's DirectInput map to convert to key codes
	if (is_extended)
		sc |= 0x80;
	uint8_t ch = LIgPlatform_ScanCodeToChar[sc];
	if (!ch)
		return;

	SetKeyState(ch, !is_up);
}

bool DDrInput_WindowProc(HWND window, UINT msg, WPARAM wparam, LPARAM lparam,
                         LRESULT* res)
{
	// This is called from our own window proc for now, so we only want to use
	// it when Daodan input is enabled
	if (!DDgUseDaodanInput)
		return false;

	switch (msg) {
		case WM_INPUT:
			if (HandleWmInput((HRAWINPUT)lparam, wparam)) {
				*res = 0;
				return true;
			}
			break;
		case WM_WINDOWPOSCHANGED:
			HandleWmWindowPosChanged((WINDOWPOS *)lparam);
			break;
		case WM_KEYDOWN:
		case WM_SYSKEYDOWN:
		case WM_KEYUP:
		case WM_SYSKEYUP:
			HandleWmKeyboard(LOWORD(wparam), LOWORD(lparam), HIWORD(lparam));
			break;
	}

	return false;
}

void ONICALL DDrActionBuffer_Get(short* count, LItActionBuffer** buffers)
{
	// So, Oni's version of this function was totally different. In unpatched
	// Oni, action buffers were produced 60 times a second by a separate
	// high-priority input thread, LIiInterruptHandleProc, and consumed by
	// LIrActionBuffer_Get, which was called from the main thread and could
	// provide multiple outstanding action buffers to the engine at once.
	//
	// That was a problem for a couple of reasons. Firstly, Windows timers
	// aren't that accurate, so even if Oni was running at a steady 60 fps, the
	// input thread would provide no input buffers or two input buffers on some
	// frames. At the default Windows timer resolution of 15ms, the input thread
	// could run as slowly as 33 Hz, which meant no input on 45% of frames.
	// Secondly, Oni seems to have problems with processing multiple input
	// buffers per-frame. Camera movement only seems to be taken from the first
	// buffer in the list.
	//
	// Maybe this design made sense for classic Windows or classic Mac OS. It's
	// also possible that it was designed this way so that complex input
	// sequences wouldn't be dropped even if Oni was dropping lots of frames.
	// Still, it seems unnecessary for modern Windows. Windows already reads
	// input for us and buffers it in the message queue. So we do something much
	// different and simpler here. We generate a single action buffer inside
	// DDrActionBuffer_Get based on the most up-to-date input.

	// Only produce input buffers when input is enabled. LIrActionBuffer_Get
	// does the same thing.
	if (!LIgMode_Internal) {
		*count = 0;
		*buffers = NULL;
		return;
	}

	// Make a copy of our temporary action buffer with all the input we've
	// gathered this frame. This is the copy that Oni's engine will see.
	static LItActionBuffer buf = { 0 };
	buf = DDgActionBuffer;
	DDgActionBuffer	= (LItActionBuffer) { 0 };

	// Add held keys to the action buffer
	for (int i = 0; i < ARRAY_SIZE(DDgInputState); i++) {
		if (DDgInputState[i]) {
			DDrActionBuffer_Add(&buf, &(LItDeviceInput) {
				.input = i,
				.analog = 1.0,
			});
		}
	}

	*count = 1;
	*buffers = &buf;
}
